package Bundle::SNMP::MIB::Compiler;

$VERSION = '0.01';

1;

__END__

=head1 NAME

Bundle::SNMP::MIB::Compiler - A bundle to install all SNMP::MIB::Compiler related modules

=head1 SYNOPSIS

C<perl -MCPAN -e 'install Bundle::SNMP::MIB::Compiler'>

=head1 CONTENTS

Data::Dumper

Data::Compare

=head1 DESCRIPTION

This bundle includes SNMP::MIB::Compiler which is a MIB compiler.

=head1 AUTHOR

Fabien Tassin E<lt>F<fta@oleane.net>E<gt>

=cut
